@@ -29,56 +29,82 @@ import shutil
 import subprocess
 import tempfile
 
-from PyQt5.QtCore import QDir, QFile, QFileInfo
+from PyQt5.QtCore import QDir, QFile, QFileInfo, QIODevice
 
-from ..project import MfsDirectory, MfsFile
+from ..project import QrcDirectory, QrcFile
 from ..user_exception import UserException
+from ..version import PYQTDEPLOY_HEXVERSION
 
 
 class Builder():
     """ The builder for a project. """
 
-    # Map PyQt modules to the corresponding qmake QT or CONFIG values.
-    class QtMetaData:
+    # Map PyQt modules to the corresponding qmake QT or CONFIG values.  A
+    # module that doesn't need any of these values can be omitted
+    class _QtMetaData:
         def __init__(self, qt=None, config=None):
             self.qt = qt
             self.config = config
 
-    _pyqt_module_map = {
-        'QAxContainer':         QtMetaData(qt=['axcontainer']),
-        'QtBluetooth':          QtMetaData(qt=['bluetooth']),
-        'QtCore':               QtMetaData(qt=['-gui']),
-        'QtDBus':               QtMetaData(qt=['dbus', '-gui']),
-        'QtDesigner':           QtMetaData(qt=['designer']),
-        'QtGui':                QtMetaData(),
-        'QtHelp':               QtMetaData(qt=['help']),
-        'QtMacExtras':          QtMetaData(qt=['macextras']),
-        'QtMultimedia':         QtMetaData(qt=['multimedia']),
-        'QtMultimediaWidgets':  QtMetaData(
+    _pyqt5_module_map = {
+        'QAxContainer':         _QtMetaData(qt=['axcontainer']),
+        'QtBluetooth':          _QtMetaData(qt=['bluetooth']),
+        'QtCore':               _QtMetaData(qt=['-gui']),
+        'QtDBus':               _QtMetaData(qt=['dbus', '-gui']),
+        'QtDesigner':           _QtMetaData(qt=['designer']),
+        'QtHelp':               _QtMetaData(qt=['help']),
+        'QtMacExtras':          _QtMetaData(qt=['macextras']),
+        'QtMultimedia':         _QtMetaData(qt=['multimedia']),
+        'QtMultimediaWidgets':  _QtMetaData(
                                         qt=['multimediawidgets',
                                                 'multimedia']),
-        'QtNetwork':            QtMetaData(qt=['network', '-gui']),
-        'QtOpenGL':             QtMetaData(qt=['opengl']),
-        'QtPositioning':        QtMetaData(qt=['positioning']),
-        'QtPrintSupport':       QtMetaData(qt=['printsupport']),
-        'QtQml':                QtMetaData(qt=['qml']),
-        'QtQuick':              QtMetaData(qt=['quick']),
-        'QtSensors':            QtMetaData(qt=['sensors']),
-        'QtSerialPort':         QtMetaData(qt=['serialport']),
-        'QtSql':                QtMetaData(qt=['sql', 'widgets']),
-        'QtSvg':                QtMetaData(qt=['svg']),
-        'QtTest':               QtMetaData(qt=['testlib', 'widgets']),
-        'QtWebKit':             QtMetaData(qt=['webKit', 'network']),
-        'QtWebKitWidgets':      QtMetaData(qt=['webkitwidgets']),
-        'QtWidgets':            QtMetaData(qt=['widgets']),
-        'QtWinExtras':          QtMetaData(qt=['winextras', 'widgets']),
-        'QtX11Extras':          QtMetaData(qt=['x11extras']),
-        'QtXmlPatterns':        QtMetaData(
+        'QtNetwork':            _QtMetaData(qt=['network', '-gui']),
+        'QtOpenGL':             _QtMetaData(qt=['opengl']),
+        'QtPositioning':        _QtMetaData(qt=['positioning']),
+        'QtPrintSupport':       _QtMetaData(qt=['printsupport']),
+        'QtQml':                _QtMetaData(qt=['qml']),
+        'QtQuick':              _QtMetaData(qt=['quick']),
+        'QtSensors':            _QtMetaData(qt=['sensors']),
+        'QtSerialPort':         _QtMetaData(qt=['serialport']),
+        'QtSql':                _QtMetaData(qt=['sql', 'widgets']),
+        'QtSvg':                _QtMetaData(qt=['svg']),
+        'QtTest':               _QtMetaData(qt=['testlib', 'widgets']),
+        'QtWebKit':             _QtMetaData(qt=['webkit', 'network']),
+        'QtWebKitWidgets':      _QtMetaData(qt=['webkitwidgets']),
+        'QtWidgets':            _QtMetaData(qt=['widgets']),
+        'QtWinExtras':          _QtMetaData(qt=['winextras', 'widgets']),
+        'QtX11Extras':          _QtMetaData(qt=['x11extras']),
+        'QtXmlPatterns':        _QtMetaData(
                                         qt=['xmlpatterns', '-gui', 'network']),
 
-        'QtChart':              QtMetaData(config=['qtcommercialchart']),
-        'QtDataVisualization':  QtMetaData(qt=['datavisualization']),
-        'Qsci':                 QtMetaData(config=['qscintilla2']),
+        'QtChart':              _QtMetaData(config=['qtcommercialchart']),
+        'QtDataVisualization':  _QtMetaData(qt=['datavisualization']),
+        'Qsci':                 _QtMetaData(config=['qscintilla2']),
+    }
+
+    _pyqt4_module_map = {
+        'QAxContainer':         _QtMetaData(config=['qaxcontainer']),
+        'QtCore':               _QtMetaData(qt=['-gui']),
+        'QtDBus':               _QtMetaData(qt=['dbus', '-gui']),
+        'QtDeclarative':        _QtMetaData(qt=['declarative', 'network']),
+        'QtDesigner':           _QtMetaData(config=['designer']),
+        'QtHelp':               _QtMetaData(config=['help']),
+        'QtMultimedia':         _QtMetaData(qt=['multimedia']),
+        'QtNetwork':            _QtMetaData(qt=['network', '-gui']),
+        'QtOpenGL':             _QtMetaData(qt=['opengl']),
+        'QtScript':             _QtMetaData(qt=['script', '-gui']),
+        'QtScriptTools':        _QtMetaData(qt=['scripttools', 'script']),
+        'QtSql':                _QtMetaData(qt=['sql']),
+        'QtSvg':                _QtMetaData(qt=['svg']),
+        'QtTest':               _QtMetaData(qt=['testlib']),
+        'QtWebKit':             _QtMetaData(qt=['webkit', 'network']),
+        'QtXml':                _QtMetaData(qt=['xml', '-gui']),
+        'QtXmlPatterns':        _QtMetaData(
+                                        qt=['xmlpatterns', '-gui', 'network']),
+        'phonon':               _QtMetaData(qt=['phonon']),
+
+        'QtChart':              _QtMetaData(config=['qtcommercialchart']),
+        'Qsci':                 _QtMetaData(config=['qscintilla2']),
     }
 
     def __init__(self, project, verbose=False):
@@ -96,23 +122,22 @@ class Builder():
 
         project = self._project
 
-        try:
-            os.makedirs(build_dir, exist_ok=True)
-        except Exception as e:
-            raise UserException("Unable to create the build directory.",
-                    str(e))
+        self._create_directory(build_dir)
 
         freeze = self._copy_lib_file('freeze.py')
 
         self._write_qmake(build_dir, freeze)
 
+        resources_dir = os.path.join(build_dir, 'resources')
+        self._create_directory(resources_dir)
+
         for resource in self.resources():
-            if resource == 'application':
+            if resource == '':
                 package = project.application_package
-                self._write_resource(build_dir, resource, package, 
+                self._write_resource(resources_dir, resource, package, 
                         project.absolute_path(package.name), freeze)
             elif resource == 'stdlib':
-                self._write_resource(build_dir, resource,
+                self._write_resource(resources_dir, resource,
                         project.stdlib_package,
                         os.path.join(project.python_target_stdlib_dir, ''),
                         freeze)
@@ -120,10 +145,11 @@ class Builder():
                 # Add the PyQt package to a temporary copy of the site-packages
                 # package.
                 site_packages_package = project.site_packages_package.copy()
+                pyqt_dir = 'PyQt5' if project.application_is_pyqt5 else 'PyQt4'
 
                 if len(project.pyqt_modules) != 0:
-                    pyqt_pkg_init = MfsFile('__init__.py')
-                    pyqt_pkg_dir = MfsDirectory('PyQt5')
+                    pyqt_pkg_init = QrcFile('__init__.py')
+                    pyqt_pkg_dir = QrcDirectory(pyqt_dir)
                     pyqt_pkg_dir.contents.append(pyqt_pkg_init)
                     site_packages_package.contents.append(pyqt_pkg_dir)
 
@@ -132,10 +158,10 @@ class Builder():
                         self._add_uic_dir(pyqt_pkg_dir,
                                 os.path.join(
                                         self._project.python_target_stdlib_dir,
-                                        'site-packages', 'PyQt5'),
+                                        'site-packages', pyqt_dir),
                                 'uic', [])
 
-                self._write_resource(build_dir, resource,
+                self._write_resource(resources_dir, resource,
                         site_packages_package,
                         os.path.join(project.python_target_stdlib_dir,
                                 'site-packages', ''),
@@ -146,7 +172,7 @@ class Builder():
     def _add_uic_dir(self, package, pyqt_dir, dirname, dir_stack):
         """ Add a uic directory to a package. """
 
-        dir_pkg = MfsDirectory(dirname)
+        dir_pkg = QrcDirectory(dirname)
         package.contents.append(dir_pkg)
 
         dirpath = [pyqt_dir] + dir_stack + [dirname]
@@ -159,7 +185,7 @@ class Builder():
             content_path = os.path.join(dirpath, content)
 
             if os.path.isfile(content_path):
-                dir_pkg.contents.append(MfsFile(content))
+                dir_pkg.contents.append(QrcFile(content))
             elif os.path.isdir(content_path):
                 dir_stack.append(dirname)
                 self._add_uic_dir(dir_pkg, pyqt_dir, content, dir_stack)
@@ -185,13 +211,12 @@ class Builder():
         qmake_qt = []
         qmake_config = []
 
+        module_map = self._pyqt5_module_map if project.application_is_pyqt5 else self._pyqt4_module_map
+
         for pyqt_m in project.pyqt_modules:
             needs_gui = True
 
-            try:
-                metadata = self._pyqt_module_map[pyqt_m]
-            except KeyError:
-                pass
+            metadata = module_map.get(pyqt_m, self._QtMetaData())
 
             if metadata.qt is not None:
                 for qt in metadata.qt:
@@ -227,10 +252,11 @@ class Builder():
         if len(project.pyqt_modules) > 0:
             sitepackages = project.absolute_path(
                     project.python_target_stdlib_dir) + '/site-packages'
+            pyqt_version = 'PyQt5' if project.application_is_pyqt5 else 'PyQt4'
 
             for pyqt in project.pyqt_modules:
                 if pyqt != 'uic':
-                    extensions['PyQt5.' + pyqt] = sitepackages + '/PyQt5'
+                    extensions[pyqt_version + '.' + pyqt] = sitepackages + '/' + pyqt_version
 
             # Add the implicit sip module.
             extensions['sip'] = sitepackages
@@ -269,6 +295,20 @@ class Builder():
 
             f.write('LIBS += -L{0} -l{1}\n'.format(lib_dir, lib))
 
+        # Add the platform specific stuff.
+        platforms_f = QFile(self._lib_filename('platforms.pro'))
+
+        if not platforms_f.open(QIODevice.ReadOnly|QIODevice.Text):
+            raise UserException(
+                    "Unable to open file {0}.".format(platforms_f.fileName()),
+                    platforms_f.errorString())
+
+        platforms = platforms_f.readAll()
+        platforms_f.close()
+
+        f.write('\n')
+        f.write(platforms.data().decode('latin1'))
+
         # Specify any resource files.
         resources = self.resources()
 
@@ -278,26 +318,29 @@ class Builder():
             f.write('RESOURCES =')
 
             for resource in resources:
-                f.write(' \\\n    {0}.qrc'.format(resource))
+                if resource == '':
+                    resource = 'pyqtdeploy'
+
+                f.write(' \\\n    resources/{0}.qrc'.format(resource))
 
             f.write('\n')
 
         # Specify the source and header files.
         f.write('\n')
 
-        f.write('SOURCES = main.c pyqtdeploy_main.c mfsimport.cpp\n')
+        f.write('SOURCES = main.c pyqtdeploy_main.c pyqtdeploy_module.cpp\n')
         self._write_main_c(build_dir, app_name, extensions.keys())
         self._copy_lib_file('pyqtdeploy_main.c', build_dir)
-        self._copy_lib_file('mfsimport.cpp', build_dir)
+        self._copy_lib_file('pyqtdeploy_module.cpp', build_dir)
 
-        f.write('HEADERS = frozen_bootstrap.h frozen_main.h\n')
+        f.write('HEADERS = frozen_bootstrap.h frozen_main.h pyqtdeploy_version.h\n')
 
         bootstrap_f = self._create_file(build_dir, '__bootstrap__.py')
         bootstrap_f.write('''import sys
-import mfsimport
+import pyqtdeploy
 
 sys.path = [{0}]
-sys.path_hooks = [mfsimport.mfsimporter]
+sys.path_hooks = [pyqtdeploy.qrcimporter]
 '''.format(', '.join(["':/{0}'".format(resource) for resource in resources])))
         bootstrap_f.close()
 
@@ -308,6 +351,12 @@ sys.path_hooks = [mfsimport.mfsimporter]
                 project.absolute_path(project.application_script), freeze,
                 main=True)
 
+        version_f = self._create_file(build_dir, 'pyqtdeploy_version.h')
+        version_f.write(
+                '#define PYQTDEPLOY_HEXVERSION %s\n' % hex(
+                        PYQTDEPLOY_HEXVERSION))
+        version_f.close()
+
         # All done.
         f.close()
 
@@ -318,8 +367,10 @@ sys.path_hooks = [mfsimport.mfsimporter]
 
         resources = []
 
-        if project.application_package.name != '':
-            resources.append('application')
+        for content in project.application_package.contents:
+            if content.included:
+                resources.append('')
+                break
 
         resources.append('stdlib')
 
@@ -333,20 +384,26 @@ sys.path_hooks = [mfsimport.mfsimporter]
 
         return resources
 
-    def _write_resource(self, build_dir, resource, package, src, freeze):
+    def _write_resource(self, resources_dir, resource, package, src, freeze):
         """ Create a .qrc file for a resource and the corresponding contents.
         """
 
-        f = self._create_file(build_dir, resource + '.qrc')
+        # The main application resource does not go in a sub-directory.
+        if resource == '':
+            qrc_file = 'pyqtdeploy.qrc'
+            dst_root_dir = resources_dir
+        else:
+            qrc_file = resource + '.qrc'
+            dst_root_dir = os.path.join(resources_dir, resource)
+            self._create_directory(dst_root_dir)
+
+        f = self._create_file(resources_dir, qrc_file)
 
         f.write('''<!DOCTYPE RCC>
 <RCC version="1.0">
     <qresource>
 ''')
 
-        dst_root_dir = os.path.join(build_dir, resource)
-        self._create_directory(dst_root_dir)
-
         src_root_dir, src_root = os.path.split(src)
 
         self._write_package_contents(package.contents, dst_root_dir,
@@ -367,17 +424,23 @@ sys.path_hooks = [mfsimport.mfsimporter]
             dir_stack = []
             dst_dir = dst_root_dir
         else:
-            dst_dir = os.path.join(dst_root_dir, dir_tail)
+            if dir_tail == '.':
+                dir_stack = []
 
-        self._create_directory(dst_dir)
+            dst_dir = os.path.join(dst_root_dir, dir_tail)
+            self._create_directory(dst_dir)
 
         prefix = os.path.basename(dst_root_dir)
+        if prefix != 'resources':
+            prefix = [prefix]
+        else:
+            prefix = []
 
         for content in contents:
             if not content.included:
                 continue
 
-            if isinstance(content, MfsDirectory):
+            if isinstance(content, QrcDirectory):
                 dir_stack.append(content.name)
                 self._write_package_contents(content.contents, dst_root_dir,
                         src_root_dir, dir_stack, freeze, f)
@@ -398,7 +461,7 @@ sys.path_hooks = [mfsimport.mfsimporter]
 
                 dst_path = os.path.join(dst_dir, dst_file)
 
-                file_path = [prefix]
+                file_path = list(prefix)
 
                 if dir_tail != '':
                     file_path.extend(dir_stack)
@@ -430,36 +493,69 @@ int main(int argc, char **argv)
         if len(extension_names) > 0:
             inittab = 'extension_modules'
 
-            for ext in extension_names:
-                base_ext = ext.split('.')[-1]
-
-                f.write('    extern PyObject *PyInit_%s(void);\n' % base_ext)
-
-            f.write('''
-    static struct _inittab %s[] = {
-''' % inittab)
-
-            for ext in extension_names:
-                base_ext = ext.split('.')[-1]
-
-                f.write('        {"%s", PyInit_%s},\n' % (ext, base_ext))
-
-            f.write('''        {NULL, NULL}
-    };
-
-''')
+            f.write('#if PY_MAJOR_VERSION >= 3\n')
+            cls._write_inittab(f, extension_names, inittab, py3=True)
+            f.write('#else\n')
+            cls._write_inittab(f, extension_names, inittab, py3=False)
+            f.write('#endif\n\n')
         else:
             inittab = 'NULL'
 
-        f.write('''    extern int pyqtdeploy_main(int argc, char **argv, wchar_t *py_main,
-            struct _inittab *extension_modules);
-
-    return pyqtdeploy_main(argc, argv, L"%s", %s);
-}
-''' % (app_name, inittab))
+        f.write('#if PY_MAJOR_VERSION >= 3\n')
+        cls._write_main_call(f, app_name, inittab, py3=True)
+        f.write('#else\n')
+        cls._write_main_call(f, app_name, inittab, py3=False)
+        f.write('#endif\n}\n')
 
         f.close()
 
+    @staticmethod
+    def _write_inittab(f, extension_names, inittab, py3):
+        """ Write the Python version specific extension module inittab. """
+
+        if py3:
+            init_type = 'PyObject *'
+            init_prefix = 'PyInit_'
+        else:
+            init_type = 'void '
+            init_prefix = 'init'
+
+        for ext in extension_names:
+            base_ext = ext.split('.')[-1]
+
+            f.write('    extern %s%s%s(void);\n' % (init_type, init_prefix,
+                    base_ext))
+
+        f.write('''
+    static struct _inittab %s[] = {
+''' % inittab)
+
+        for ext in extension_names:
+            base_ext = ext.split('.')[-1]
+
+            f.write('        {"%s", %s%s},\n' % (ext, init_prefix, base_ext))
+
+        f.write('''        {NULL, NULL}
+    };
+''')
+
+    @staticmethod
+    def _write_main_call(f, app_name, inittab, py3):
+        """ Write the Python version specific call to pyqtdeploy_main(). """
+
+        if py3:
+            name_type = 'wchar_t'
+            name_prefix = 'L'
+        else:
+            name_type = 'char'
+            name_prefix = ''
+
+        f.write('''    extern int pyqtdeploy_main(int argc, char **argv, %s *py_main,
+            struct _inittab *extension_modules);
+
+    return pyqtdeploy_main(argc, argv, %s"%s", %s);
+''' % (name_type, name_prefix, app_name, inittab))
+
     def _freeze(self, output, py_filename, freeze, main=False, as_data=False):
         """ Freeze a Python source file to a C header file or a data file. """
 
@@ -472,16 +568,29 @@ int main(int argc, char **argv)
         args.append(output)
         args.append(py_filename)
 
-        self._log("Freezing {0}".format(py_filename))
+        self._log("Running '{0}'".format(' '.join(args)))
 
         try:
-            subprocess.check_output(args, stderr=subprocess.STDOUT)
+            subprocess.check_output(args, stderr=subprocess.STDOUT,
+                    universal_newlines=True)
         except subprocess.CalledProcessError as e:
+            self._log(e.output)
             raise UserException("Unable to freeze {0}.".format(py_filename),
                     e.output)
 
     @staticmethod
-    def _copy_lib_file(filename, dirname=None):
+    def _lib_filename(filename):
+        """ Get name of a file in the 'lib' directory that can be used by
+        QFile.
+        """
+
+        lib_dir = QFileInfo(__file__).dir()
+        lib_dir.cd('lib')
+
+        return lib_dir.filePath(filename)
+
+    @classmethod
+    def _copy_lib_file(cls, filename, dirname=None):
         """ Copy a library file to a directory and return the full pathname of
         the copy.  If the directory wasn't specified then copy it to a
         temporary directory.
@@ -497,9 +606,7 @@ int main(int argc, char **argv)
         d_filename = os.path.join(dirname, filename)
 
         # The source filename.
-        s_dir = QFileInfo(__file__).dir()
-        s_dir.cd('lib')
-        s_filename = s_dir.filePath(filename)
+        s_filename = cls._lib_filename(filename)
 
         # Make sure the destination doesn't exist.
         QFile.remove(d_filename)
@@ -526,9 +633,11 @@ int main(int argc, char **argv)
         """ Create a directory which may already exist. """
 
         try:
-            os.mkdir(dir_name)
-        except FileExistsError:
-            pass
+            os.makedirs(dir_name, exist_ok=True)
+        except Exception as e:
+            raise UserException(
+                    "Unable to create the '{0}' directory.".format(dir_name),
+                    str(e))
 
     def _log(self, message):
         """ Log a message if requested. """
