@@ -34,6 +34,7 @@ from PyQt5.QtWidgets import (QFileDialog, QMainWindow, QMessageBox, QTabWidget,
 from ..builder import Builder
 from ..project import Project
 from ..user_exception import UserException
+from ..version import PYQTDEPLOY_RELEASE
 
 from .application_page import ApplicationPage
 from .extension_modules_page import ExtensionModulesPage
@@ -122,6 +123,7 @@ class ProjectGUI(QMainWindow):
         menu_bar.addSeparator()
 
         help_menu = menu_bar.addMenu("&Help")
+        help_menu.addAction("About pyqtdeploy...", self._about)
         help_menu.addAction(QWhatsThis.createAction(help_menu))
 
     def _create_central_widget(self):
@@ -129,12 +131,38 @@ class ProjectGUI(QMainWindow):
 
         tabs = QTabWidget()
 
-        for page_factory in (ApplicationPage, PyQtPage, StdlibPage, SitePackagesPage, ExtensionModulesPage, PythonPage):
-            page = page_factory()
-            tabs.addTab(page, page.label)
+        application_page = ApplicationPage()
+        tabs.addTab(application_page, application_page.label)
+
+        pyqt_page = PyQtPage()
+        tabs.addTab(pyqt_page, pyqt_page.label)
+
+        stdlib_page = StdlibPage()
+        tabs.addTab(stdlib_page, stdlib_page.label)
+
+        site_packages_page = SitePackagesPage()
+        tabs.addTab(site_packages_page, site_packages_page.label)
+
+        extension_modules_page = ExtensionModulesPage()
+        tabs.addTab(extension_modules_page, extension_modules_page.label)
+
+        python_page = PythonPage()
+        tabs.addTab(python_page, python_page.label)
+
+        application_page.pyqt_version_changed.connect(
+                pyqt_page.set_pyqt_version)
 
         self.setCentralWidget(tabs)
 
+    def _about(self):
+        """ Tell the user about the application. """
+
+        QMessageBox.about(self, "About pyqtdeploy",
+"""This is pyqtdeploy v%s
+
+pyqtdeploy is a tool for deploying PyQt4 and PyQt5 applications written using Python v2.6 or later or Python v3.3 or later to desktop and mobile devices.
+""" % PYQTDEPLOY_RELEASE)
+
     def _new_project(self):
         """ Create a new, unnamed project. """
 
@@ -277,7 +305,7 @@ class ProjectGUI(QMainWindow):
 
         settings = QSettings()
 
-        self.resize(settings.value('size', QSize(600, 400)))
+        self.resize(settings.value('size', QSize(860, 400)))
         self.move(settings.value('pos', QPoint(200, 200)))
 
     def _save_settings(self):
